#
# Agate commands main entry point
#
import argparse
import getpass
import sys

import obiba_agate.rest as rest
import obiba_agate.auth_management.application as application
import obiba_agate.auth_management.group as group
import obiba_agate.auth_management.user as user

def prompt_password():
    return getpass.getpass(prompt='Enter password:')

def add_agate_arguments(parser):
    """
    Add Agate access arguments
    """
    parser.add_argument('--agate', '-ag', required=False, default='http://localhost:8081',
                        help='Agate server base url (default: http://localhost:8081)')
    parser.add_argument('--user', '-u', required=False, help='User name')
    parser.add_argument('--password', '-p', nargs="?", required=False, help='User password')
    parser.add_argument('--otp', '-ot', action='store_true', help='Whether a one-time password is to be provided (required when connecting with username/password AND two-factor authentication is enabled)')
    parser.add_argument('--ssl-cert', '-sc', required=False, help='Certificate (public key) file')
    parser.add_argument('--ssl-key', '-sk', required=False, help='Private key file')
    parser.add_argument('--verbose', '-v', action='store_true', help='Verbose output')
    parser.add_argument('--no-ssl-verify', '-nv', action='store_true', help='Do not verify SSL certificates for HTTPS.')


def add_subcommand(subparsers, name, help, add_args_func, default_func):
    """
    Make a sub-parser, add default arguments to it, add sub-command arguments and set the sub-command callback function.
    """
    subparser = subparsers.add_parser(name, help=help)
    add_agate_arguments(subparser)
    add_args_func(subparser)
    subparser.set_defaults(func=default_func)


def run():
    """
    Command-line entry point.
    """
    # Parse arguments
    parser = argparse.ArgumentParser(description='Agate command line.')
    subparsers = parser.add_subparsers(title='sub-commands',
                                      help='Available sub-commands. Use --help option on the sub-command '
                                            'for more details.')

    # Add subcommands
    add_subcommand(subparsers, 'rest', 'Request directly the Agate REST API, for advanced users.', rest.add_arguments,
                  rest.do_command)
    add_subcommand(subparsers, 'add-user', 'Add a new user.',
                  user.user_add_arguments, user.do_add_command)
    add_subcommand(subparsers, 'delete-user', 'Delete a user.',
                  user.user_delete_arguments,
                  user.do_delete_command)
    add_subcommand(subparsers, 'add-group', 'Add a new group.',
                  group.group_add_arguments,
                  group.do_add_command)
    add_subcommand(subparsers, 'delete-group', 'Delete a group (does NOT delete users of the group).',
                  group.group_delete_arguments,
                  group.do_delete_command)
    add_subcommand(subparsers, 'add-application', 'Add a new application.',
                  application.application_add_arguments,
                  application.do_add_command)
    add_subcommand(subparsers, 'delete-application',
                  'Delete an application.',
                  application.application_delete_arguments,
                  application.do_delete_command)

    # Execute selected command
    args = parser.parse_args()

    try:
        if not args.password or len(args.password) == 0:
            args.password = prompt_password()
            pass
    except AttributeError:
        args.password = prompt_password()

    if hasattr(args, 'func'):
        try:
            args.func(args)
        except Exception as e:
            print(e)
            sys.exit(2)
    else:
      print('Agate command line tool.')
      print('For more details: agate --help')



"""
Python commands for adding or deleting a Group
"""

import json
import obiba_agate.core as core


def group_add_arguments(parser):
    """
    Add agate group management arguments
    """

    parser.add_argument('--name', help='The group Name (required), it must be unique', required=True)
    parser.add_argument('--description', help='The group Description', required=False)
    parser.add_argument('--applications',
                        help='Members of a group get access to the applications associated to this group',
                        required=False, nargs='*')


def do_add_command(args):
    """
    Execute add group management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    if args.verbose:
        request.verbose()

    group = {'name': args.name}
    if args.description:
        group['description'] = args.description
    if args.applications:
        group['applications'] = args.applications

    request.post().content_type_json().resource(core.UriBuilder(['groups']).build()).content(
        json.dumps(group))

    response = request.send()
    print(response.content)

def group_delete_arguments(parser):
    """
    Add agate group management arguments
    """

    parser.add_argument('--name', help='The group Name (required)', required=True)


def do_delete_command(args):
    """
    Execute delete group management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    request.delete().content_type_json().resource(core.UriBuilder(['group', args.name]).build())

    response = request.send()
    print(response.content)



"""
Generic Agate web services command.
"""

import sys
import obiba_agate.core as core


def add_arguments(parser):
    """
    Add REST command specific options
    """
    parser.add_argument('ws', help='Web service path, for instance: /study/xxx')
    parser.add_argument('--method', '-m', required=False,
                        help='HTTP method (default is GET, others are POST, PUT, DELETE, OPTIONS)')
    parser.add_argument('--accept', '-a', required=False, help='Accept header (default is application/json)')
    parser.add_argument('--content-type', '-ct', required=False,
                        help='Content-Type header (default is application/json)')
    parser.add_argument('--json', '-j', action='store_true', help='Pretty JSON formatting of the response')


def do_command(args):
    """
    Execute REST command
    """
    # Build and send request
    client = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args))

    try:
        request = client.new_request()
        request.fail_on_error()

        if args.accept:
            request.accept(args.accept)
        else:
            request.accept_json()

        if args.content_type:
            request.content_type(args.content_type)
            print('Enter content:')
            request.content(sys.stdin.read())

        if args.verbose:
            request.verbose()

        # send request
        request.method(args.method).resource(args.ws)
        response = request.send()

        # format response
        res = response.content
        if args.json:
            res = response.pretty_json()
        elif args.method in ['OPTIONS']:
            res = response.headers['Allow']

        # output to stdout
        print(res)
    finally:
        client.close()


"""
Based on Python Request library https://docs.python-requests.org/en/latest/index.html
"""

import base64
import json
import os.path
import getpass
from requests import Session, Request
import urllib3
from http.client import HTTPConnection
from http import HTTPStatus
import urllib.parse
from functools import reduce


class AgateClient:
    """
    AgateClient holds the configuration for connecting to Agate.
    """

    def __init__(self, server=None):
        self.session = Session()
        self.headers = {}
        self.base_url = self.__ensure_entry('Agate address', server)

    def __del__(self):
        self.close()

    @classmethod
    def build(cls, loginInfo):
        return AgateClient.buildWithAuthentication(loginInfo.data['server'], loginInfo.data['user'],
                                                   loginInfo.data['password'], loginInfo.data['otp'],
                                                   loginInfo.data["no_ssl_verify"])

    @classmethod
    def buildWithAuthentication(cls, server, user, password, otp=None, no_ssl_verify: bool = False):
        client = cls(server)
        if client.base_url.startswith('https:'):
            client.session.verify = not no_ssl_verify
            if no_ssl_verify:
                urllib3.disable_warnings()

        client.credentials(user, password, otp)
        return client

    def __ensure_entry(self, text, entry, pwd=False):
        e = entry
        if not entry:
            if pwd:
                e = getpass.getpass(prompt=text + ': ')
            else:
                e = input(text + ': ')
        return e

    def credentials(self, user, password, otp):
        u = self.__ensure_entry('User name', user)
        p = self.__ensure_entry('Password', password, True)
        if otp:
            val = input("Enter 6-digits code: ")
            self.header('X-Obiba-TOTP', val)

        self.session.headers.update({"Authorization": "Basic %s" % base64.b64encode(("%s:%s" % (u, p)).encode("utf-8")).decode("utf-8")})

    def verify(self, value):
        """
        Ignore or validate certificate

        :param value = True/False to validation or not. Value can also be a CA_BUNDLE file or directory (e.g. 'verify=/etc/ssl/certs/ca-certificates.crt')
        """
        self.session.verify = value
        return self

    def header(self, key, value):
        """
        Adds a header to session headers used by the request

        :param key - header key
        :param value - header value
        """
        header = {}
        header[key] = value

        self.session.headers.update(header)
        return self

    def new_request(self):
        return AgateRequest(self)

    def close(self):
        """
        Closes client session and request to close Agate server session
        """
        try:
            self.new_request().resource("/auth/session/_current").delete().send()
            self.session.close()
        except Exception as e:
            pass

    class LoginInfo:
        data = None

        @classmethod
        def parse(cls, args):
            data = {}
            argv = vars(args)

            data["no_ssl_verify"] = argv.get("no_ssl_verify")

            if argv.get('agate'):
                data['server'] = argv['agate']
            else:
                raise Exception('Agate server information is missing.')

            if argv.get('user') and argv.get('password'):
                data["user"] = argv["user"]
                data["password"] = argv["password"]
                data["otp"] = argv["otp"]
            else:
                raise Exception('Invalid login information. Requires user-password or certificate-key information')

            setattr(cls, 'data', data)
            return cls()

        def isSsl(self):
            if self.data.keys() & {'cert', 'key'}:
                return True
            return False


class AgateRequest:
    """
    Agate request.
    """

    def __init__(self, agate_client):
        self.client = agate_client
        self.options = {}
        self.headers = {'Accept': 'application/json'}
        self._verbose = False
        self.params = {}
        self._fail_on_error = False
        self.files = None
        self.data = None


    def timeout(self, value):
        """
        Sets the connection and read timeout
        Note: value can be a tupple to have different timeouts for connection and reading (connTimout, readTimeout)

        :param value - connection/read timout
        """
        self.options["timeout"] = value
        return self

    def verbose(self):
        """
        Enables the verbose mode
        """
        HTTPConnection.debuglevel = 1
        self._verbose = True
        return self

    def fail_on_error(self):
        self._fail_on_error = True
        return self

    def header(self, key, value):
        """
        Adds a header to session headers used by the request

        :param key - header key
        :param value - header value
        """
        if value:
            header = {}
            header[key] = value
            self.headers.update(header)
        return self

    def accept(self, value):
        self.headers.update({"Accept": value})
        return self

    def content_type(self, value):
        return self.header('Content-Type', value)

    def accept_json(self):
        return self.accept('application/json')

    def content_type_json(self):
        self.content_type('application/json')
        return self

    def method(self, method):
        """
        Sets a HTTP method

        :param method - any of ['GET', 'DELETE', 'PUT', 'POST', 'OPTIONS']
        """
        if not method:
            self.method = "GET"
        elif method in ["GET", "DELETE", "PUT", "POST", "OPTIONS"]:
            self.method = method
        else:
            raise ValueError("Not a valid method: " + method)
        return self

    def get(self):
        return self.method('GET')

    def put(self):
        return self.method('PUT')

    def post(self):
        return self.method('POST')

    def delete(self):
        return self.method('DELETE')

    def options(self):
        return self.method('OPTIONS')

    def __build_request(self):
        """
        Builder method creating a Request object to be sent by the client session object
        """
        request = Request()
        request.method = self.method if self.method else "GET"

        for option in self.options:
            setattr(request, option, self.options[option])

        # Combine the client and the request headers
        request.headers = {}
        request.headers.update(self.client.session.headers)
        request.headers.update(self.headers)

        if self.resource:
            path = self.resource
            request.url = self.client.base_url + "/ws" + path

            if self.params:
                request.params = self.params
        else:
            raise ValueError("Resource is missing")

        if self.files is not None:
            request.files = self.files

        if self.data is not None:
            request.data = self.data

        return request


    def resource(self, ws):
        self.resource = ws
        return self

    def form(self, parameters):
        """
        Stores the request's body as a form
        Note: no need to transform parameters in key=value pairs

        :param parametes - parameters as a dict value
        """
        return self.content(parameters)

    def content(self, content):
        """
        Stores the request body

        :param content - request body
        """
        if self._verbose:
            print("* Content:")
            print(content)

        self.data = content
        return self


    def content_upload(self, filename):
        """
        Sets the file associate with the upload

        Note: Requests library takes care of mutlti-part setting in the header
        """
        if self._verbose:
            print("* File Content:")
            print("[file=" + filename + ", size=" + str(os.path.getsize(filename)) + "]")
        with open(filename, "rb") as file:
            self.files = {"file": (filename, file.read())}
        return self

    def send(self):
        """
        Sends the request via client session object
        """
        request = self.__build_request()
        response = AgateResponse(self.client.session.send(request.prepare()))

        if self._fail_on_error and response.code >= 400:
            raise HTTPError(response)

        return response


class Storage:
    """
    Content storage.
    """

    def __init__(self):
        self.content = ''
        self.line = 0

    def store(self, buf):
        self.line = self.line + 1
        self.content = self.content + buf.decode("utf-8")

    def __str__(self):
        return self.contents


class HeaderStorage(Storage):
    """
    Store response headers in a dictionary: key is the header name,
    value is header value or the list of header values.
    """

    def __init__(self):
        Storage.__init__(self)
        self.headers = {}

    def store(self, buf):
        Storage.store(self, buf)
        header = buf.decode("utf-8").partition(':')
        if header[1]:
            value = header[2].rstrip().strip()
            if header[0] in self.headers:
                current_value = self.headers[header[0]]
                if isinstance(current_value, str):
                    self.headers[header[0]] = [current_value]
                self.headers[header[0]].append(value)
            else:
                self.headers[header[0]] = value


class AgateResponse:
    """
    Response from Agate: code, headers and content
    """

    def __init__(self, response):
        self.response = response

    @property
    def code(self):
        return self.response.status_code

    @property
    def headers(self):
        return self.response.headers

    @property
    def content(self):
        return self.response.content

    @property
    def version(self):
        return self.headers.get("X-Agate-Version", None)

    @property
    def version_info(self):
        agateVersion = self.version
        if agateVersion is not None:
            info = {}
            version_parts = self.version.split(".")
            if len(version_parts) == 3:
                info["major"], info["minor"], info["patch"] = version_parts
                return info
            else:
                # Handle malformed version string
                return None
        return None

    def as_json(self):
        """
        Returns response body as a JSON document
        """
        if self.response is None or self.response.content is None:
            return None

        try:
            return self.response.json()
        except Exception as e:
            if type(self.response.content) == str:
                return self.response.content
            else:
                # FIXME silently fail
                return None

    def pretty_json(self):
        """
        Beatifies the JSON response
        """
        return json.dumps(self.as_json(), sort_keys=True, indent=2)


class UriBuilder:
    """
    Build a valid Uri.
    """

    def __init__(self, path=[], params={}):
        self.path = path
        self.params = params

    def path(self, path):
        self.path = path
        return self

    def segment(self, seg):
        self.path.append(seg)
        return self

    def params(self, params):
        self.params = params
        return self

    def query(self, key, value):
        self.params.update([(key, value),])
        return self

    def __str__(self):
        def concat_segment(p, s):
            return p + '/' + s

        def concat_params(k):
            return urllib.parse.quote(k) + '=' + urllib.parse.quote(str(self.params[k]))

        def concat_query(q, p):
            return q + '&' + p

        p = urllib.parse.quote('/' + reduce(concat_segment, self.path))
        if len(self.params):
            q = reduce(concat_query, list(map(concat_params, list(self.params.keys()))))
            return p + '?' + q
        else:
            return p

    def build(self):
        return self.__str__()

class HTTPError(Exception):
    """
    HTTP related error class
    """

    def __init__(self, response: AgateResponse, message: str = None):
        # Call the base class constructor with the parameters it needs
        super().__init__(message if message else "HTTP Error: %s" % response.code)
        self.code = response.code
        http_status = [x for x in list(HTTPStatus) if x.value == response.code][0]
        self.message = message if message else "%s: %s" % (http_status.phrase, http_status.description)
        self.error = response.as_json() if response.content else {"code": response.code, "status": self.message}
        # case the reported error is not a dict
        if type(self.error) != dict:
            self.error = {"code": response.code, "status": self.error}

    def is_client_error(self) -> bool:
        return self.code >= 400 and self.code < 500

    def is_server_error(self) -> bool:
        return self.code >= 500


"""
Python commands for adding or deleting an Application
"""

import json
import obiba_agate.core as core


def application_add_arguments(parser):
    """
    Add agate application management arguments
    """

    parser.add_argument('--name', help='The application Name (required), it must be unique', required=True)
    parser.add_argument('--description', help='The application description', required=False)
    parser.add_argument('--key', help='The application Key (required)', required=True)
    parser.add_argument('--redirect', help='Callback URL to the application\'s server, required in the OAuth context',
                        required=False)
    parser.add_argument('--manual-approval', help='User signing up through the application is in "pending for approval" state', required=False)


def do_add_command(args):
    """
    Execute add application management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    if args.verbose:
        request.verbose()

    application = {'name': args.name, 'key': args.key}
    if args.description:
        application['description'] = args.description
    if args.redirect:
        application['redirectURI'] = args.redirect
    if args.manual_approval:
        application['autoApproval'] = False

    request.post().content_type_json().resource(core.UriBuilder(['applications']).build()).content(
        json.dumps(application))

    response = request.send()
    print(response.content)


def application_delete_arguments(parser):
    """
    Add agate application management arguments
    """

    parser.add_argument('--name', help='The application Name (required)', required=True)


def do_delete_command(args):
    """
    Execute delete application management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    request.delete().content_type_json().resource(core.UriBuilder(['application', args.name]).build())

    response = request.send()
    print(response.content)



"""
Python commands for adding or deleting a User
"""

import json
import obiba_agate.core as core


def user_add_arguments(parser):
    """
    Add agate user management arguments
    """

    parser.add_argument('--name', help='The User Name (required), it must be unique', required=True)
    parser.add_argument('--email', help='The Email (required), it must be unique', required=True)
    parser.add_argument('--upassword', help='The user Password (required if realm is "agate-user-realm")', required=False)
    parser.add_argument('--first-name', help='The user First Name', required=False)
    parser.add_argument('--last-name', help='The user Last Name', required=False)
    parser.add_argument('--realm', help='Realm in which the user will authenticate (default is: "agate-user-realm")',
                        required=False, default='agate-user-realm')
    parser.add_argument('--applications', help='Applications in which the user can sign in, e.g. "opal mica"',
                        required=False, nargs='*')
    parser.add_argument('--groups',
                        help='Members of a group get access to the applications associated to this group, e.g. \
                        "mica-administrator mica-reviewer mica-editor mica-data-access-officer mica-user \
                        opal-administrator"',
                        required=False, nargs='*')
    parser.add_argument('--role', help='A simple user can only access to its own account (default is: "agate-user")',
                        required=False, default='agate-user')
    parser.add_argument('--status', help='Only active users can sign in (default is "ACTIVE")', required=False,
                        default='ACTIVE')


def do_add_command(args):
    """
    Execute add user management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    if args.verbose:
        request.verbose()

    user = {'name': args.name, 'email': args.email, 'role': args.role, 'status': args.status,
            'realm': args.realm}
    if args.first_name:
        user['firstName'] = args.first_name
    if args.last_name:
        user['lastName'] = args.last_name
    if args.applications:
        user['applications'] = args.applications
    if args.groups:
        user['groups'] = args.groups

    if not args.upassword and args.realm == 'agate-user-realm':
        raise Exception("User password is required for Agate's realm.")

    data = {'user': user}
    if args.realm == 'agate-user-realm':
        data = {'password': args.upassword, 'user': user}

    request.post().content_type_json().resource(core.UriBuilder(['users']).build()).content(json.dumps(data))

    response = request.send()
    print(response.content)


def user_delete_arguments(parser):
    """
    Add agate user management arguments
    """

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--name', help='User name')
    group.add_argument('--email', help='User email')


def do_delete_command(args):
    """
    Execute delete user management command
    """

    request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    request.fail_on_error()

    request.get().resource(
        core.UriBuilder(['users', 'find']).query('q', args.name if args.name else args.email).build())

    user_response = request.send()

    new_request = core.AgateClient.build(core.AgateClient.LoginInfo.parse(args)).new_request()
    new_request.fail_on_error()

    new_request.delete().resource(core.UriBuilder(['user', json.loads(user_response.content)['id']]).build())
    response = new_request.send()

    print(response.content)
